#ifndef WINSYSTEM_SYSTEM_H
#define WINSYSTEM_SYSTEM_H

#include <vector>
#include <string>

#include "math/vec.h"
#include "system/gl.h"
#include "system/system.h"

class WinSystem : public PSystem {
public:
	WinSystem() {reset(); }
	~WinSystem() {shutdown(); }

	virtual void init();
	virtual void shutdown();
	
	virtual void loop(void (*loopCallback)());
	virtual void requestExit();

	virtual void setInitDoubleBuffered(bool db);
	virtual void setInitPos(const vec2i &initPos);
	virtual void setInitSize(const vec2i &initSize);

	virtual const std::string &getArg(int index) const;
	virtual size_t getArgCount() const;

	HWND getHWnd() { return hwnd; }

protected:
//set before init.  kind of a messy system =) I'll work it out -- put an init structure in or something ... no I won't.
	bool initFullScreen;
	bool initDBuf;
	vec2i initPos;
	vec2i initSize;
	int initBPP;
	const char *initTitle;

	std::vector<std::string> args;

	//this needs to be public 'cuz of the GLUT using methods as callbacks, etc
	void doLoop();

	//for all those who want to swap/update display buffers manually inter-render
	void updateDisplayBuffers();
	
	void reset();

	//internal loop methods
	//for now I'm going to use a callback method... maybe switch it to an (abstract?) object later...
	void (*loopCallback)();

	void preUpdate();
	void postUpdate();
	
	//stored state variables
	bool dbuf;
	bool exitRequested;

	//system-specific
	HWND hwnd;				//current window
	HDC hDC;				//device context
	HGLRC hRC;				//gl context
	DEVMODE devmodeBackup;	//stored mode when we started
};

#endif	//WINSYSTEM_SYSTEM_H
