#include <iostream>	//include this before stdio or it complains about exit being defined twice
#include <stdio.h>	//NULL
#include "system/memory.h"	//the system/ prefix means 'we gotta pull this from the generic-system code'
#include "system/singletons.h"
#include "system/exception.h"
#include "winsystem/system.h"
#include "winsystem/timer.h"
#include "winsystem/singletons.h"
#include "winsystem/wstring.h"

using namespace std;

//////////////////////////////// keyboard system mapping

//the header file had 127 or so virtual keys.  maybe 256 maybe 512
//someone emailed me about an out of bounds virtual key value: 15xx
//if Windows has any more keys than this, I might have to resort to using stl maps ...
static Key::Key_t keyForVirtKey[2048];

static void initKeyMaps() {

	for (int i = 0; i < numberof(keyForVirtKey); i++) {
		keyForVirtKey[i] = (Key::Key_t)-1;
	}

#define VKBRIDGE2(x,y)			keyForVirtKey[VK_##x] = Key::y;
#define VKBRIDGE(x)				keyForVirtKey[VK_##x] = Key::x; 
#define VKBRIDGERANGE(x,y)		for (int i = VK_##x; i <= VK_##y; i++) keyForVirtKey[i] = (Key::Key_t)(i - VK_##x + Key::x);
#define VKBRIDGERANGE2(x,y,z)	for (int i = VK_##x; i <= VK_##y; i++) keyForVirtKey[i] = (Key::Key_t)(i - VK_##x + Key::z);
	for (int i = 0; i < 127; i++) {
		if (!strlen(Key::name[i])) continue;	//if we're not mapping the key anyways then skip it, huh?
		int vk = VkKeyScan(i);
		if (vk > numberof(keyForVirtKey)) {
			//what this says is there exists virtual keycodes that windows could return which are larger than
			// our keyForVirtKey map above.
			//maybe I should just use a stl map rather than allocating this huge array?  that would save on space
			// but would slow down our lookup time from constant- to logarithm.
			//meh, it's only windows.  not like things don't run slow enough already.
			cerr << "required virtual key index " << vk << " for key " << i << "/" << (char)i << "." << endl
				<< "I need to up the maximum keyForVirtKey array to include this size" << endl;
			exit(0);
		}
		keyForVirtKey[vk] = (Key::Key_t)i;
	}
	VKBRIDGE2(DELETE,DEL)
	VKBRIDGERANGE(F1, F15)
	//windows wont do two buttons... i tihnk the l and r shfit keys are there so ppl can get frustrated like I did ...
	VKBRIDGE(LSHIFT) VKBRIDGE(RSHIFT)
	VKBRIDGE2(SHIFT,LSHIFT)
	VKBRIDGE2(LCONTROL,LCTRL) VKBRIDGE2(RCONTROL,RCTRL)
	VKBRIDGE2(CONTROL,LCTRL)
	VKBRIDGE2(LMENU,LALT)
	VKBRIDGE2(RMENU,RALT)
	VKBRIDGERANGE2(NUMPAD0,NUMPAD9,KP0)
	VKBRIDGE(UP)
	VKBRIDGE(DOWN)
	VKBRIDGE(LEFT)
	VKBRIDGE(RIGHT)
	VKBRIDGE2(CAPITAL,CAPSLOCK)
	VKBRIDGE2(LWIN, LMETA)
	VKBRIDGE2(RWIN, RMETA)
}

//////////////////////////////// class code

void WinSystem::reset() {
	//init data values here
	hwnd = NULL;
	hDC = NULL;
	hRC = NULL;
	//devmode anyone?

	loopCallback = NULL;
	exitRequested = false;
	
	initSize.set(720,480);
	initPos.set(	//depends on initSize
		(GetSystemMetrics(SM_CXSCREEN) >> 1) - (initSize.x >> 1),
		(GetSystemMetrics(SM_CYSCREEN) >> 1) - (initSize.y >> 1));
	initBPP = 0;
	initFullScreen = false;
	initDBuf = true;
	initTitle = "Windows Window";
}

static LONG WINAPI MyWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	static PAINTSTRUCT ps;

	switch(uMsg) {
	case WM_PAINT:
		BeginPaint(hWnd, &ps);
		EndPaint(hWnd, &ps);
		return 0;

	case WM_SIZE:
		disp.resize(LOWORD(lParam),HIWORD(lParam));
#if 0	//if we ever want window position data ...
		if (GetWindowRect(hWnd, &displayH->displayRect)) {
			displayH->displayCenter.x = (displayH->displayRect.left + displayH->displayRect.right) >> 1;
			displayH->displayCenter.y = (displayH->displayRect.top + displayH->displayRect.bottom) >> 1;
		}
#endif
		return 0;

	case WM_CLOSE:
		sys.requestExit();
		return 0;

// ********* MOUSE WINDOWS MESSAGES ********* 

//dirty of me...
#ifndef WM_MOUSEWHEEL	//???
#define WM_MOUSEWHEEL                   0x020A
#endif
		///all the same deal
	case WM_MOUSEWHEEL:
		mouse.setButtonPress(mouse.BUTTON_LEFT,		!!(wParam & MK_LBUTTON));
		mouse.setButtonPress(mouse.BUTTON_MIDDLE,	!!(wParam & MK_MBUTTON));
		mouse.setButtonPress(mouse.BUTTON_RIGHT,	!!(wParam & MK_RBUTTON));
		mouse.setCursorPos(LOWORD(lParam),HIWORD(lParam));
		mouse.accumDZ(HIWORD(wParam));	//scroll
		return 0;

	case WM_MOUSEMOVE:
	case WM_LBUTTONUP:
	case WM_MBUTTONUP:
	case WM_RBUTTONUP:
	case WM_LBUTTONDOWN:
	case WM_MBUTTONDOWN:
	case WM_RBUTTONDOWN:
		mouse.setButtonPress(mouse.BUTTON_LEFT,		!!(wParam & MK_LBUTTON));
		mouse.setButtonPress(mouse.BUTTON_MIDDLE,	!!(wParam & MK_MBUTTON));
		mouse.setButtonPress(mouse.BUTTON_RIGHT,	!!(wParam & MK_RBUTTON));
		mouse.setCursorPos(LOWORD(lParam),HIWORD(lParam));
		return 0;

// ********* KEYBOARD WINDOWS MESSAGE ********* 

	case WM_KEYDOWN:
//		printf("key:%d\n", wParam);
		assert(wParam >= 0 && wParam < numberof(keyForVirtKey));
		kbd.keyDown(keyForVirtKey[wParam]);
		return 0;
	case WM_KEYUP:
		assert(wParam >= 0 && wParam < numberof(keyForVirtKey));
		kbd.keyUp(keyForVirtKey[wParam]);
		return 0;

#if 0
// ********* CD AUDIO WINDOWS MESSAGE ********* 
	case WM_DEVICECHANGE:
		if (cdAudio) {
			cdAudio->DeviceChange((unsigned int)wParam, (unsigned int)lParam);
			return 0;
		}
		break;
	case MM_MCINOTIFY:				//for some odd reason this is only coming through after i move the window
		if (cdAudio) return cdAudio->SendMessage(wParam, lParam);
		break;
// ********* CALLBACK (?) *********
	case WM_MENUSELECT:
		if (main_display->menuSelectCallback) {
			main_display->menuSelectCallback((UINT)LOWORD(wParam), (UINT)HIWORD(wParam), (HMENU)lParam);
			return 0;
		}
		break;
	case WM_COMMAND:
		if (main_display->commandCallback) {
			if (main_display->commandCallback(HIWORD(wParam), LOWORD(wParam), (HWND)lParam)) {
				return 0;
			}
		}
		break;
#endif
	}
#if 0
	if (main_display->messageCallback) {
		if (main_display->messageCallback(hWnd, uMsg, wParam, lParam)) {
			return 0;
		}
	}
#endif
	return (LONG)DefWindowProc(hWnd, uMsg, wParam, lParam); 
} 

void WinSystem::init() {

	//store local parameters
	dbuf = initDBuf;

	memSystemInit();

	//init our timer whenever
	timer.init();

// BEGIN WINDOWS INIT register our window

	try {
		wstring wndClassName = str2wstr("WinSystemWindowClass");

		HINSTANCE hInst = GetModuleHandle(NULL);
		if (!hInst) throw Exception() << "failed to get HINSTANCE";

		WNDCLASS wc;
		memset(&wc, 0, sizeof(wc));
		wc.style         = CS_OWNDC;
		wc.lpfnWndProc   = MyWndProc;
		wc.cbClsExtra    = 0;
		wc.cbWndExtra    = 0;
		wc.hInstance     = hInst;
		wc.hIcon         = LoadIcon(NULL, IDI_WINLOGO);
		wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
		wc.hbrBackground = NULL;
		wc.lpszMenuName  = NULL;
		wc.lpszClassName = wndClassName.c_str();

		if (!RegisterClass(&wc)) throw Exception() << "RegisterClass() failed";

		DWORD dwExStyle =	WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
		DWORD dwStyle =		WS_OVERLAPPEDWINDOW | WS_CLIPSIBLINGS | WS_CLIPCHILDREN;

		wstring initTitleW = str2wstr(initTitle);

		hwnd = CreateWindowEx(dwExStyle, wndClassName.c_str(), initTitleW.c_str(), dwStyle, VEC2ELEM(initPos), VEC2ELEM(initSize), NULL, NULL, hInst, NULL);
		if (!hwnd) throw Exception() << "failed to create window";

		hDC = GetDC(hwnd);
		if (!hDC) throw Exception() << "GetDC failed";

		PIXELFORMATDESCRIPTOR pfd;
		memset(&pfd, 0, sizeof(pfd));
		int iPixelFormat = GetPixelFormat(hDC);			 		//get our current format
		DescribePixelFormat(hDC, iPixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &pfd);				//modify it

		pfd.nSize		= sizeof(pfd);
		pfd.nVersion	= 1;
		pfd.dwFlags		= PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL;
		if (dbuf) pfd.dwFlags |= PFD_DOUBLEBUFFER;
		pfd.iPixelType	= PFD_TYPE_RGBA;
		pfd.iLayerType	= PFD_MAIN_PLANE;
		//FIXME: stencil RW stutters like mad when run in 16 bit color - not HW accelerated? whats up?
		pfd.cColorBits = 32;
		pfd.cAlphaBits = 8;
		pfd.cStencilBits = 16;

		int pf = ChoosePixelFormat(hDC, &pfd);								//find the best fit pixel format to set it to
		if (!pf) throw Exception() << "ChoosePixelFormat failed";
		if (!SetPixelFormat(hDC, pf, &pfd)) throw Exception() << "SetPixelFormat failed";		//and set it accordingly

		hRC = wglCreateContext(hDC);												//make the GL RC
		if (!hRC) throw Exception() << "wglCreateContext failed";
		if (!wglMakeCurrent(hDC, hRC)) throw Exception() << "wglMakeCurrent failed";		//and make it the current RC

		//ResetViewport
		//set up the init viewport size
		RECT r;
		if (!GetClientRect(hwnd, &r)) throw Exception() << "GetClientRect failed";
		initSize.set(r.right, r.bottom);		//should we store this here or in a sep variable?

		//good to go - show the window and leave
		ShowWindow(hwnd, SW_SHOW);

		//initialize fullscreen as false - and save current DEVMODE info
		EnumDisplaySettings(NULL, ENUM_CURRENT_SETTINGS, &devmodeBackup);
		initFullScreen = false;

	} catch (Exception &t) {
		throw Error() << "WinSystem::init failed! " << t;
	}

//while i'm here lets store the command line vector ...
	int nargs = 0;
	LPWSTR *argv = CommandLineToArgvW(GetCommandLineW(), &nargs);
	if (!argv) {
		throw Error() << "WinSystem::init failed to get command line";
	} else {
		for (int i = 0; i < nargs; i++) {
			size_t len = wcslen(argv[i])+1;
			char *str = new char[len+1];
			assert(str);
			memset(str, 0, len+1);
			for (int j = 0; j < len; j++) {
				str[j] = (char)argv[i][j];
			}
			args.push_back(str);
			delete str;
		}
		LocalFree(argv);
	}
	
	//looks like sdl doesnt set an initial video resize message ...
	disp.resize(VEC2ELEM(initSize));

	//keyboard stuff
	initKeyMaps();
}

void WinSystem::shutdown() {
	//shutdown windows
	wglMakeCurrent(NULL, NULL);
	if (hwnd) ReleaseDC(hwnd, hDC);
	if (hRC) wglDeleteContext(hRC);

	//shutdown any objects *HERE*
	memSystemShutdown();

	//and reset...
	reset();
}

void WinSystem::setInitDoubleBuffered(bool db) { initDBuf = db; }
void WinSystem::setInitPos(const vec2i &_initPos) { initPos = _initPos; }
void WinSystem::setInitSize(const vec2i &_initSize) { initSize = _initSize; }

const std::string &WinSystem::getArg(int index) const { return args[index]; }
size_t WinSystem::getArgCount() const { return args.size(); }

void WinSystem::requestExit() {
	exitRequested = true;
}

void WinSystem::loop(void (*lc)()) {
	MSG msg;
	//store the main loop's callback method
	loopCallback = lc;

	while (!exitRequested) {
		if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
			if (msg.message == WM_QUIT) break;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		} else {
			winSys.doLoop();
		}
	}
}

void WinSystem::doLoop() {
	preUpdate();
	if (loopCallback) loopCallback();
	postUpdate();
}

void WinSystem::preUpdate() {
	//system update init
	disp.update();	//update resolution
	mouse.update();
	kbd.preUpdate();
}

void WinSystem::postUpdate() {
	//system update shutdown
	kbd.postUpdate();
	//NOTICE - does this apply to all, or only to the local viewport?
	updateDisplayBuffers();
}

void WinSystem::updateDisplayBuffers() {
	if (dbuf) {
		SwapBuffers(hDC);
	} else {
		glFlush();
	}
}
